import java.io.*;
import java.net.*;
import java.util.*;

/**
*	This class creates the initial physical and virtual root objects for the
*	FTP service, reads and parses the general configuration file, and does some
*	basic/background management for the GUI. The FTPServer object also accepts
*	new client connections and creates FTPConnection objects for them.
*/
public class FTPServer extends Thread {
	//Static, final, identifying variables used to display the startup banner
	public static final String TITLE = "Wrabbit FTP Server";
	public static final String VERSION = "Version Alpha 5";
	public static final String COPYRIGHT = "Copyright (c) 2001 - Charles R Berube";
	public static final String INFORMATION1 = "http://www.orderlychaos.com";
	public static final String INFORMATION2 = "http://sourceforge.net/project/?group_id=4728";

	//Connection and network related variables
	private ServerSocket server;
	private Socket incoming;
	private Vector activeConnections;

	//Internal server status and file system information
	private boolean running = false;
	private FTPObject rootObject;

	//Server settings and configuration files
	private String configurationFile = new String("config");
	private String virtualRootsFile = new String("virtualroots");
	private int port = 21;

	//GUI variables
	private MainWindow mainWindow;
	private CurrentConnectionsPanel connectionPanel;

	/**
	*	Construct an instance of the FTPServer object.
	*/
	public FTPServer() {
		activeConnections = new Vector();
		connectionPanel = null;
		rootObject = new FTPObject("root", "root", 0, true, null);
	}

	/**
	*	Complete the creation and setup of the FTPServer object.
	*	Needs to be seperate to support command line parameters well.
	*/
	public void initialize() {
		getConfiguration();

		addFileRoots(rootObject);
		addVirtualRoots(rootObject);

		try {
			server = new ServerSocket(port);
		} catch (Exception e) {
			System.out.println("[S] startup> Server creation failed.");
			System.out.println("    java error> " + e);
		}
	}

	/**
	*	Parse the virtualroots file and try to attach a virtual root to
	*	each path listed in it. Prints a notice, but does not crash out
	*	if a virtual root cannot be established.
	*/
	private void addVirtualRoots(FTPObject target) {
		BufferedReader in = null;
		String line = null, vroot = null, realpath = null;
		StringTokenizer st = null;

		try {
			in = new BufferedReader(new FileReader(new File(virtualRootsFile)));
		} catch (Exception nfe) {
			System.out.println("[S] startup> Cannot read virtual roots file.");
		}

		try {
			while ( (line = in.readLine()) != null ) {
				if (!line.trim().equalsIgnoreCase("") && !(line.startsWith(";"))) {
					vroot = line;
					line = in.readLine();
					realpath = line;

					if (new File(realpath).exists()) {
						target.addChild(new FTPObject(vroot, new File(realpath), target));
					} else {
						System.out.println("[S] startup> Unable to create virtual root linked to");
						System.out.println("             '" + realpath + "'.");
					}
				}
			}

			in.close();
		} catch (Exception e) {
			System.out.println("[S] startup> Error reading virtual roots file");
		}
	}

	/**
	*	Discovers file system roots, and adds them to the 'target' object.
	*	TODO: Support linux/unix roots.
	*/
	private void addFileRoots(FTPObject target) {
		StringTokenizer driveList = new StringTokenizer("C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z", ",", false);
		String thisDrive;
		File attempt;

		while (driveList.hasMoreTokens() == true) {
			thisDrive = driveList.nextToken();
			try {
				attempt = new File(thisDrive + ":\\");
				if (attempt.isDirectory()) {
					target.addChild(new FTPObject("Local Disk " + thisDrive, attempt, target));
				}
			} catch (Exception e) {
				System.out.println(e);
			}
		}
	}

	/**
	*	Waits for and accepts client connections.
	*/
	public void run() {
		FTPConnection newConnection;
		byte[] connectionCheck = new byte[7];

		running = true;

		System.out.println("[S] startup> Configuration file: " + configurationFile);
		System.out.println("[S] startup> Port: " + server.getLocalPort());

		while (running) {
			try {
				incoming = server.accept();

				//Check to see if this connection is from the localhost, and if so,
				//try and read a string from it immediately to see if it matches
				//the string 'showgui'.
				if (incoming.getInetAddress().equals(InetAddress.getLocalHost())) {
					incoming.getInputStream().read(connectionCheck, 0, 7);
					if ((new String(connectionCheck)).equals("showgui")) {
						mainWindow.show();
						incoming.close();
						continue;
					}
				}

				newConnection = new FTPConnection(this, incoming, rootObject, rootObject);
				newConnection.start();
				activeConnections.add(newConnection);
				notifyConnectionPanel();
			} catch (Exception e) {
				System.out.println("Incoming connection listening failed");
				System.out.println(e);
			}
		}
	}

	/**
	*	Returns true if the server thread is running
	*/
	public boolean isRunning() {
		return running;
	}

	/**
	*	Read the configuration file, and parse out the settings from it.
	*/
	private void getConfiguration() {
		BufferedReader in = null;
		String line = null, parameter = null, setting = null;
		StringTokenizer st = null;

		try {
			in = new BufferedReader(new FileReader(new File(configurationFile)));
		} catch (Exception nfe) {
			System.out.println("Cannot read configuration file, defaulting as many settings as possible");
			System.out.println(nfe);
		}

		try {
			while ( (line = in.readLine()) != null ) {
				st = new StringTokenizer(line, ":", false);
				parameter = st.nextToken();

				//Port setting
				if (parameter.equalsIgnoreCase("port")) {
					setting = st.nextToken();
					port = new Integer(setting).intValue();
				} else

				//Virtual roots file setting
				if (parameter.equalsIgnoreCase("virtualrootsfile")) {
					setting = st.nextToken();
					virtualRootsFile = setting;
				}
			}

			in.close();
		} catch (Exception e) {
			System.out.println("Error reading configuration file, using defaults for all settings that could not be read");
		}
	}

	/**
	*	Return the server's root object (the top of the directory
	*	tree).
	*/
	public FTPObject getRoot() {
		return rootObject;
	}

	/**
	*	Return the vector of active connections.
	*/
	public Vector getActiveConnections() {
		return activeConnections;
	}

	/**
	*	Called by an active FTPConnection object just as it becomes
	*	inactive, normally, or due to an error.
	*/
	public void signalConnectionTerminated(FTPConnection connection) {
		activeConnections.remove(connection);
		notifyConnectionPanel();
	}

	/**
	*	If a connection panel has been registered with the server,
	*	notify it that it is time to refresh the connections list.
	*/
	private void notifyConnectionPanel() {
		if (connectionPanel != null) {
			connectionPanel.fillConnectionList();
		}
	}

	public void refreshConnectionPanel() {
		connectionPanel.repaint();
	}

	/**
	*	Register a connection panel with the FTPServer.
	*/
	public void registerConnectionPanel(CurrentConnectionsPanel p) {
		connectionPanel = p;
	}

	/**
	*	Application entry point. Accepts command line arguments:
	*	-gui: to display the GUI and start the server
	*	-guionly: which displays the GUI and does NOT start the server
	*	-config <filename>: specify a filename for the configuration file
	*/
	public static void main(String[] args) {
		boolean showGUI, runServer;
		int currentArg;

		Thread serverThread;
		FTPServer ftp;

		//Display the startup banner
		System.out.println(FTPServer.TITLE);
		System.out.println(FTPServer.VERSION);
		System.out.println(FTPServer.COPYRIGHT);
		System.out.println("");
		System.out.println(FTPServer.INFORMATION1);
		System.out.println(FTPServer.INFORMATION2);
		System.out.println("");

		//Setup initial parameters
		runServer = true;
		showGUI = false;

		//Create an instance of the FTPServer object
		ftp = new FTPServer();

		//Parse command line parameters if there are any
		for (currentArg=0;currentArg<args.length;currentArg++) {
			//-gui
			if (args[currentArg].equalsIgnoreCase("-gui")) {
				showGUI = true;
			}

			//-guionly
			if (args[currentArg].equalsIgnoreCase("-guionly")) {
				showGUI = true;
				runServer = false;
			}

			//-config
			if (args[currentArg].equalsIgnoreCase("-config")) {
				if (currentArg == args.length-1) {
					System.out.println("[S] command line> No configuration filename specified with the -config");
					System.out.println("                  parameter; the default file, 'config', will be used.");
				} else {
					ftp.configurationFile = args[++currentArg];
				}
			}
		}

		//Initialize the server -- this must be done here because of the alterable configuration file
		ftp.initialize();

		//Always build the main window, even if we aren't going to show it.
		ftp.mainWindow = new MainWindow(ftp);

		//Maybe show the GUI window
		if (showGUI) {
			ftp.mainWindow.pack();
			ftp.mainWindow.show();
		}

		//Maybe run the server, or display a message and wait for the window to close
		if (runServer) {
			ftp.start();
		} else {
			System.out.println("Running in GUI mode only, FTP service not started.");
			//while (ftp.mainWindow.isVisible() == false) { }
			//try { while (ftp.mainWindow.isVisible() == true) { Thread.currentThread().sleep(500); } } catch (Exception ex) { }
		}
	}
}
